/*
	The hCount algorithm
	
	As described in  - 
	@inproceedings{refHC,
		author = {Cheqing Jin and Weining Qian and Chaofeng Sha and Jeffrey X. Yu and Aoying Zhou},
		title = {Dynamically maintaining frequent items over a data stream},
		booktitle = {CIKM '03: Proceedings of the twelfth international conference on Information and knowledge management},
		year = {2003},
		isbn = {1-58113-723-0},
		pages = {287--294},
		location = {New Orleans, LA, USA},
		doi = {http://doi.acm.org/10.1145/956863.956918},
		publisher = {ACM Press},
		address = {New York, NY, USA},
	}
*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <values.h>

#include "hc.h"

#define	INSERT	1
#define DELETE	-1

/* macro that returns a random number in [v1, v2) */
#define random(v1,v2) (((float)(v1))+(((float)(v2)-((float)(v1)))*rand()/(RAND_MAX+1.0f)))

/* 
Initialize the hCount algorithm
*/
hC_type *hC_Init(float epsilon, float delta, long M)
{
    float e;			/* min x/ln(x) and x > 1  --> evaluates to the base of the natural logarithm */
    int R_RANGE;		/* range for random numbers a_i and b_i */
    int i;
    S_TYPE *aptr;
	float rho = 1-delta;

    hC_type *result = calloc(1, sizeof(hC_type));
    if (result == NULL) {
		printf("\ncalloc error");
		exit(100);
    }

    result->N = 0;			/* total transactions */
    result->M = M; 
    e = 2.718281828;
    result->P = 1048583;	/* big Prime number */

    R_RANGE = result->M;
    result->m =  (int) M_E / epsilon;		/* length of hash table */
    result->h = (int) (log(-result->M / log(rho)));		/* width of hash table */
    //printf("\nHC*: m=%d, h=%d", result->m, result->h);

	/* Allocate memory for S */
	aptr = (S_TYPE *) malloc(result->m * result->h * sizeof(S_TYPE));
    if (aptr == NULL) {
		printf("\nmalloc() error.");
		exit(100);
    }
    result->S = (S_TYPE **) malloc(result->m * sizeof(S_TYPE));
    if (result->S == NULL) {
		printf("\nmalloc() error.");
		exit(200);
    }
    for (i = 0; i < result->m; i++) {
		result->S[i] = aptr + i * result->h;
    }
    
	/* Set all counters to 0 */
	memset((void *) aptr, 0,
	       result->m * result->h * sizeof(S_TYPE) / sizeof(char));
    
	/* generate h pairs of random numbers a_i and b_i  */
	result->a = (int *) malloc(result->h * sizeof(int));
    if (result->a == NULL) {
		printf("\nmalloc() error.");
		exit(300);
    }
    result->b = (int *) malloc(result->h * sizeof(int));
    if (result->b == NULL) {
		printf("\nmalloc() error.");
		exit(400);
    }
    
	for (i = 0; i < result->h; i++) {
		result->a[i] = random(0, R_RANGE);
		result->b[i] = random(0, R_RANGE);
    }

	return (result);
}

/*
Update the hCount table with new transaction
k = item
ttype = INSERT/DELETE
*/
void hC_Update(hC_type * hc, int k, int ttype)
{
    int pos, j;

    if (ttype == INSERT) {
		hc->N++;
    } else {
		hc->N--;
    }
    
	for (j = 0; j < hc->h; j++) {
		pos = ((((long long) hc->a[j] * k + hc->b[j]) % hc->P) % hc->m);

		if (ttype == INSERT) {
			hc->S[pos][j] = hc->S[pos][j] + 1;
		} else {
			hc->S[pos][j] = hc->S[pos][j] - 1;
		}
    }
}


/* 
Output all items with threshhold greater than th
*/
unsigned int *hC_Output(hC_type * hc, int th, int STAR)
{
    int k, c, j, min, pos;
    unsigned int *list;
    int res = 0;		/* counter for no. of freq. items */
	int correction = 0;	// for hCount*
	
    list = calloc(hc->N / th, sizeof(unsigned int));
    if (list == NULL) {
		printf("\nCalloc error. ");
		exit(400);
    }
 
 	if (STAR) {
	 	for (k = hc->M+1; k <= hc->M+20; k++) {
	 		min = MAXINT;
			for (j = 0; j < hc->h; j++) {
				pos = ((((long long) hc->a[j] * k + hc->b[j]) % hc->P) % hc->m);
				c = hc->S[pos][j];
				if (c < min) {
					min = c;
				}
			}
			correction += min;
	 	}
	 	correction = (int)(correction/20.0);
	}
	 	
	for (k = 0; k < hc->M; k++) {
		min = MAXINT;
		for (j = 0; j < hc->h; j++) {
			pos = ((((long long) hc->a[j] * k + hc->b[j]) % hc->P) % hc->m);
			c = hc->S[pos][j];
			if (c < min) {
				min = c;
			}
		}
		
		min -= correction;		// hCount* correction
		
		if (min > th) {
			res++;
			list[res] = k;
		}
    }

    list[0] = res;
    return (list);
}


/*
Be nice and free used memory
*/
void hC_Destroy(hC_type * hc)
{
    free(hc->a);
    free(hc->b);
    free(hc->S);
	free(hc);
}

/*
Returns the space used up by hCount
*/
int hC_Size(hC_type * hc)
{
    int size;
    size = hc->m * hc->h * sizeof(S_TYPE) + 2 * hc->h * sizeof(int) + sizeof(hC_type);
    return (size);
}


/*
Mostly for debugging
Prints relevant information about the hCount data structure
*/
void hC_Print(hC_type * hc)
{

    int i, j;

    for (i = 0; i < hc->m; i++) {
	printf("\n");
	for (j = 0; j < hc->h; j++)
	    printf("%d ", hc->S[i][j]);
    }

    for (i = 0; i < hc->h; i++) {
	printf("\n%d, a=%d, b=%d", i, hc->a[i], hc->b[i]);
    }

    printf("\nm = %d", hc->m);
    printf("\nh = %d", hc->h);
    printf("\nP = %d", hc->P);

    printf("\nhC_Print -- END");
    getchar();

}
